/*
 *  Copyright (C) 2020 Jens Thoms Toerring <jt@toerring.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 *  The program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "kv_map.h"
#include <string.h>
#include <stdlib.h>


/* Minimum number of buckets */

#define MIN_BUCKET_CNT   8


/* The number of buckets is dynamically adjusted to reflect the numberr of
 * nodes in use. For higher speed we want quite a number of buckets in
 * excess of the number of nodes. Unforturnately, as a rule of thumb,
 * the number of buckets would have to be about the square of the number
 * of nodes to have any reasonable chance of avoiding lots of collisions.
 * This would require huge number of buckets already for medium sized
 * node counts, so we have to accept relatively high collisions counts to
 * keep memory consumption in check.
 *
 * When the ratio of nodes to buckets exceeds this value double the number
 * of buckets. */

#define LOAD_FACTOR_INCR      0.7

/* When the ratio of nodes to buckets drops below this value halve the
 * number of buckets. This typically should be significantly below half
 * of the load factor for increases. Otherwise you end up with a "jo-jo"
 * effect where the number of buckets is in- and decreased on each node
 * insertion and removal, which takes quite a lot of time. */

#define LOAD_FACTOR_DECR      0.15


/* After the node structure the object stored in the node and its key
 * will be placed (as indicated by the commented out code) */

typedef struct node {
    struct node * next;              // pointer to next node
    uint32_t      hash;              // hash value of the node
//  T             object;            // object to be stored
//  const char    key[];             // key of the node
} node_t;


/*---------------------------------------------------------------*//*
 * @brief Dan J. Bernstein's favourite hash function
 *
 * @param  s  string to be hashed
 *
 * @returns an uin32_t with has value of the string
 *//*---------------------------------------------------------------*/

static uint32_t
djb2(const char * s)
{
    unsigned int h = 5381;

    while (s && *s)
        h = ((h << 5) + h) ^ *s++;       // h_i = 33 * h_{i - 1} ^ s_{i}

    return h;
}


/*---------------------------------------------------------------*//*
 * @brief Returns key of the given node
 *//*---------------------------------------------------------------*/

static inline const char *
node_to_key(const kv_map_t * map,
            const node_t   * node)
{
    return node ? (const char *) node + map->koffset : NULL;
}


/*---------------------------------------------------------------*//*
 * @brief Returns pointer to the value of the given node
 *//*---------------------------------------------------------------*/

static inline void *
node_to_value(const kv_map_t * map,
              const node_t   * node)
{
    return node ? (char *) node + map->voffset : NULL;
}


/*---------------------------------------------------------------*//*
 * @brief Returns pointer to the bucket the given hash value is in
 *//*---------------------------------------------------------------*/

static inline node_t **
bucket_ptr(const kv_map_t * map,
           uint32_t         hash)
{
    return (node_t **) map->buckets + hash % map->bucket_cnt;
}


/*---------------------------------------------------------------*//*
 * @bried Returns pointer to the first node in the map
 *//*---------------------------------------------------------------*/

static node_t *
find_first_node(const kv_map_t * map)
{
    node_t ** buckets = map->buckets;
    size_t pos = 0;

    /* Find first non-empty bucket */

    while (pos < map->bucket_cnt && ! buckets[pos])
        pos++;

    /* Return first node in the first non-empty bucket */

    return pos < map->bucket_cnt ? buckets[pos] : NULL;
}


/*---------------------------------------------------------------*//*
 * @brief Allocates memory for a new node and initializes it
 *//*---------------------------------------------------------------*/

static node_t *
create_node(kv_map_t   * map,
            const char * key)
{
	/* The first time this function is called for a map find a suitable
     * offset for the object to be stored and the key, to be stored after
     * the node. The address for the object comes as tightly after the hash
     * value in the mode as possible but must be aligned correctly. After
     * the object the key will be placed (as a char array it has no special
     * alignment requirements). */

    if (! map->voffset) {
        map->voffset = offsetof(node_t, hash) + sizeof(uint32_t);
        if (map->voffset % map->align) {
            map->voffset =   ((map->voffset + map->align) / map->align)
                           * map->align;
        }

        map->koffset = map->voffset + map->obj_size;
    }

    size_t klen = strlen(key) + 1;

	/* Allocate all the memory needed */

    node_t * node = malloc(map->koffset + klen);
    if (! node)
        return NULL;

	/* Initialize the "administrative part of the new node. */

    node->next  = NULL;
    node->hash  = djb2(key);

	/* Store the key and initialize the value part, using a user supplied
     * function if available (if that function returns non-zero throw away
     * the new node). Otherwize zero the object out. */

    strcpy((char *) node_to_key(map, node), key);

    if (! map->init) {
        memset(node_to_value(map, node), 0, map->obj_size);
    } else if (map->init(node_to_key(map, node), node_to_value(map, node))) {
        free(node);
        return NULL;
    }

    map->mem_size += map->koffset + klen;
    return node;
}


/*---------------------------------------------------------------*//*
 * @brief Returns a pointer to the node with the given key
 *        (or NULL if no node with the key exists).
 *//*---------------------------------------------------------------*/

static node_t *
find_node(kv_map_t   * map,
          const char * key)
{
    if (! map->buckets)
        return NULL;

	/* Find the proper bucket for the hash of the key */

    uint32_t hash = djb2(key);
    node_t * node = *bucket_ptr(map, hash);

    /* Iterate over the nodes in the linked list of the bucket until the
     * sought-for node has been found or we reached the end. Beside com-
     * paring the hashes also the keys are compared to guard against
     * (unlikely) hash collisions of the hash value obtained from djb2(). */

    while (node) {
        if (node->hash == hash && ! strcmp(node_to_key(map, node), key))
            break;
        node = node->next;
    }

    return node;
}


/*---------------------------------------------------------------*//*
 * @brief Inserts a node into the map
 *//*---------------------------------------------------------------*/

static void
insert_node(kv_map_t * map,
            node_t   * node)
{
	/* Find the bucket for the node from its hash */

    node_t ** bp = bucket_ptr(map, node->hash);

	/* Insert it as the first element of the linked list of nodes in the
     * bucket. */

    if (*bp)
        node->next = *bp;
    else
        node->next = NULL;

    *bp = node;
    map->node_cnt++;
}


/*---------------------------------------------------------------*//*
 * @brief Used during resizing of the bucket array: makes a linked
 *        list out of all the nodes.
 *
 * @returns the head of the linked list (or NULL if there are no nodes)
 *//*---------------------------------------------------------------*/

static node_t *
chain_nodes(kv_map_t * map)
{
    if (! map->node_cnt)
        return NULL;

    node_t * nodes = find_first_node(map);
    if (! nodes)
        return NULL;

    node_t ** buckets = map->buckets;
    size_t pos = bucket_ptr(map, nodes->hash) - (node_t**) map->buckets;

    node_t * node = nodes;

    while (node) {
        /* Find last node in the bucket */

        if (node->next) {
            node = node->next;
            continue;
        }

        /* Find the next non-empty bucket */

        while (++pos < map->bucket_cnt && ! buckets[pos])
            /* empty */ ;

        if (pos == map->bucket_cnt)
            break;

        /* Make the first element of the new bucket the successor of the
         * last node of the previous non-empty bucket. Rinse and repeat
         * until we've gone throuuig all nodes. */

        node = node->next = buckets[pos];
    }

    return nodes;
}


/*---------------------------------------------------------------*//*
 * @bried Called after resizing the bucket list to redistribute the
 *        nodes, converted to a single linked list in chain_nodes(),
 *        back into the linked lists of the buckets.
 *//*---------------------------------------------------------------*/

static void
reinsert_nodes(kv_map_t * map,
               node_t   * nodes)
{
    while (nodes) {
        node_t * next = nodes->next;
        insert_node(map, nodes);
        nodes = next;
    }
}


/*---------------------------------------------------------------*//*
 * @brief Resizes the bucket list of the map
 *//*---------------------------------------------------------------*/

static int
resize_map(kv_map_t * map,
           size_t     new_size)
{
    /* Always keep a minimum number of buckets */

    if (new_size < MIN_BUCKET_CNT)
        new_size = MIN_BUCKET_CNT;

    if (new_size == map->bucket_cnt)
        return 0;

    /* Create a single linked list out of the nodes in the linked lists of
	 * the buckets. */

    node_t * nodes = chain_nodes(map);

    /* Reallocate memory for the buckets - failure isn't a deadly issue
	 * (unless on the very first allocation) as we still can work with a
	 * smaller number of buckets, but look-ups will be slower. */

    node_t ** buckets = realloc(map->buckets, new_size * sizeof *buckets);

    if (! map->buckets && ! buckets)         // that's really bad (but unlikely)
        return -1;

    if (buckets) {
        map->mem_size += (new_size - map->bucket_cnt) * sizeof *buckets;
        map->buckets    = buckets;
        map->bucket_cnt = new_size;
    }

    /* Redistribute the nodes again over the new buckets, currently organized
     * into a single linked list. */

    memset(buckets, 0, map->bucket_cnt * sizeof *buckets);
    map->node_cnt = 0;

    reinsert_nodes(map, nodes);

    return 0;
}


/*---------------------------------------------------------------*//*
 * @brief Removes a node from the list and frees it.
 *//*---------------------------------------------------------------*/

static void
remove_node(kv_map_t   * map,
            const char * key,
            bool         disable_resize)
{
    node_t * node = find_node(map, key);
    if (! node)
        return;

    node_t ** bp = bucket_ptr(map, node->hash);

	/* If the node to remove is the first in the bucket set the bucket
	 * pointer to it's successor. Otherwise remove it from the linked list
     * of the bucket. */

    if (node == *bp) {
        *bp = node->next;
    } else {
        node_t * cur_node = *bp;
        if (cur_node->next != node)
            cur_node = cur_node->next;
        cur_node->next = node->next;
    }

    map->node_cnt--;

    /* If defined call a user supplied destructor for the value of the node,
     * then deallocate it. */

    if (map->del)
        map->del(node_to_key(map, node), node_to_value(map, node));

    map->mem_size -= map->koffset + strlen(node_to_key(map, node)) + 1;

    free(node);

	/* If, by removing the node, the number of nodes has dropped below a
	 * certain fraction of the number of buckets reduce the number of
	 * buckets. */

    if (   ! disable_resize
        && map->bucket_cnt > MIN_BUCKET_CNT
        && LOAD_FACTOR_DECR * map->bucket_cnt > map->node_cnt)
        resize_map(map, map->bucket_cnt / 2);
}


/*---------------------------------------------------------------*//*
 * @brief Returns the node a iterator is currently set to.
 *//*---------------------------------------------------------------*/

static node_t *
iter_to_node(kv_map_iter_t * iter)
{
    if (! iter || ! iter->map || ! iter->node)
        return NULL;

	/* In a freshly initialized iterator the map and node pointers are
	 * identical (that's how we know that it hasn't been used yet). Set
	 * the node pointer to the first node in the map. */

    if (iter->node == iter->map)
        iter->node = find_first_node(iter->map);

    return iter->node;
}


/*---------------------------------------------------------------*//*
 * @brief Sets up an iterator for accessing the next
 *        element of its map.
 *//*---------------------------------------------------------------*/

static void
inc_iter(kv_map_iter_t * iter)
{
    node_t * node = iter_to_node(iter);
    if (! node)
        return;

	/* If the node the iterator is set to has a successor set the iterator
     * node to it and we're done */

    if (node->next) {
        iter->node = node->next;
        return;
    }

	/* Otherwise search for the next non0empty bucket and set the iterators
     * node to its first node (or to NULL if nothing is left). */

    node_t **bp = bucket_ptr(iter->map, node->hash);
    size_t pos = bp - (node_t **) iter->map->buckets;

    while (++pos < iter->map->bucket_cnt && ! *++bp)
        /* empty */ ;

    if (pos < iter->map->bucket_cnt)
        iter->node = *bp;
    else
        iter->node = NULL;
}


/*========================================================*
 *       API functions
 *========================================================*/

/*---------------------------------------------------------------*//*
 * @brief Returns if a key-value pair for the given
 *        key exists in the map.
 *
 * @param  map  map to check in
 * @param  key  key to check for
 *
 * @returns 'true' if the key-value pair with the given
 *          key exists in the map, oterwise 'false'.
 *//*---------------------------------------------------------------*/

bool
kv_map_exists(kv_map_t   * map,
              const char * key)
{
    if (! map || ! key || ! map->buckets)
        return false;

    return find_node(map, key);
}


/*---------------------------------------------------------------*//*
 * @brief Creates a new key-valie pair in the map and
 *        returns a pointer to the value.
 *
 * Note: if the map was initialized to automatically
 * create new key-value pairs if the value of a non-
 * existing key-value pair was requested this function
 * isn't needed, but can be called without introducing
 * problems.
 *
 * If a key-value pair for the given key already exists
 * no new node is created but a pointer to the value of
 * the already existing key-value pair is returned, i.e.,
 * the call is the equivalent to one of kv_map_get().
 *//*---------------------------------------------------------------*/

void *
kv_map_add(kv_map_t   * map,
           const char * key)
{
    if (! map || ! key)
        return NULL;

    node_t * node = find_node(map, key);
    if (node)
        return node_to_value(map, node);

    /* Note: a failure to resize the map is only possible when the map is
     * completely empty */

    if (   map->node_cnt + 1 > LOAD_FACTOR_INCR * map->bucket_cnt
        && resize_map(map, map->node_cnt? 2 * map->bucket_cnt : MIN_BUCKET_CNT)
                                                                          == -1)
        return NULL;

    node = create_node(map, key);
    if (! node)
        return NULL;

    insert_node(map, node);

    return node_to_value(map, node);
}


/*---------------------------------------------------------------*//*
 * @brief Removes the key-value pair for a key
 *//*---------------------------------------------------------------*/

void
kv_map_remove(kv_map_t   * map,
              const char * key)
{
    remove_node(map, key, false);
}


/*---------------------------------------------------------------*//*
 * @brief Returns a pointer to the value for the given key.
 *
 * @param  map  map the key-value pair is in
 * @param  key  key of the value to be returned
 *
 * Note: depending on hpw the map was initialized the function
 * either returns NULL if no key-value pair was found for the
 * given key, or a pointer to the value of a newly alocated node.
 *
 * @returns poiner to value for the key or, if the map isn;t set
 *          to auto-create, NULL if no node wih the key exists.
 *//*---------------------------------------------------------------*/

void *
kv_map_get(kv_map_t   * map,
           const char * key)
{
	if (! map || ! key || (! map->buckets && ! map->auto_create))
        return NULL;

    node_t * node = find_node(map, key);
    if (! node && map->auto_create)
        return kv_map_add(map, key);

    return node_to_value(map, node);
}


/*---------------------------------------------------------------*//*
 * @brief Returns the next key from the map the iterator was
 *        initialized for.
 *
 * @param  iter  iterator for the map
 *
 * Note: if the end of the map has been reached (in which case
 * NULL is returned) the iterator jumos back to the start of
 * the map.
 *
 * @returns pointer to the next key or NULL when the end of the
 *          map has been reached.
 *//*---------------------------------------------------------------*/

const char *
kv_map_next_key(kv_map_iter_t * iter)
{
    const char * key = node_to_key(iter->map, iter_to_node(iter));

    if (key)
        inc_iter(iter);
    else
        iter->node = (node_t *) iter->map;

    return key;
}


/*---------------------------------------------------------------*//*
 * @brief Returns a pointer to the value from the map
 *        the iterator was initialized for.
 *
 * @param  iter  iterator for the map
 *
 * Note: if the end of the map has been reached (in which case
 * NULL is returned) the iterator jumos back to the start of
 *
 * @returns pointer to the next value or NULL when the end of
 *          the map has been reached.
 *//*---------------------------------------------------------------*/

void *
kv_map_next_value(kv_map_iter_t * iter)
{
    void * value = node_to_value(iter->map, iter_to_node(iter));

    if (value)
        inc_iter(iter);
    else
        iter->node = (node_t *) iter->map;

    return value;
}


/*---------------------------------------------------------------*//*
 * @brief Returns the next key and a pointer to the value from
 *        the map the iterator was initialized for.
 *
 * @param  iter   iterator to be used
 * @param  key_p    pointer to string pointer for returning the kety
 * @param  value_p  pointer for returning pointer to the value
 *
 * Note: if the iterator already had reached the end of the map
 * (in which case 0 is returned) key_p and value_p are unchanged
 * and the iterator jumps back to the start of the map.
 *
 * @returs true if a key-value pair was available, othewise false
 *//*---------------------------------------------------------------*/

bool
kv_map_next(kv_map_iter_t  * iter,
            const char    ** key_p,
            void          ** value_p)
{
    node_t * node = iter_to_node(iter);

    if (node) {
        *key_p   = node_to_key(iter->map, node);
        *value_p = node_to_value(iter->map, node);
        inc_iter(iter);
    } else {
        iter->node = (node_t *) iter->map;
    }

    return node;
}


/*---------------------------------------------------------------*//*
 * @brief Frees all memory allocated for the map
 *
 * @param  map  the map to clean out
 *//*---------------------------------------------------------------*/

void
kv_map_clear(kv_map_t * map)
{
    if (! map)                    // map does not exist
        return;

    if (! map->buckets)           // map is already empty
        return;

    kv_map_iter_t * iter = & KV_MAP_ITER_INIT(map);
    const char * key;
    while ((key = kv_map_next_key(iter)))
        remove_node(map, key, true);

    free(map->buckets);
    map->buckets    = NULL;
    map->bucket_cnt = 0;
    map->node_cnt   = 0;
    map->mem_size   = 0;
}


/*---------------------------------------------------------------*//*
 * @brief Returns the number of key-value pairs stored in the map
 *
 * @param  map  pointer to the map
 *
 * @returns number of nodes in the map
 *//*---------------------------------------------------------------*/

size_t
kv_map_count(const kv_map_t * map)
{
    return map ? map->node_cnt : 0;
}


/*---------------------------------------------------------------*//*
 * @brief Returns the number of buckets currently allocated
 *
 * @param  map  pointer to the map
 *
 * @returns current number of buckets
 *//*---------------------------------------------------------------*/

size_t
kv_map_bucket_count(const kv_map_t * map)
{
    return map ? map->bucket_cnt : 0;
}


/*---------------------------------------------------------------*//*
 * @brief Returns the number of buckets not in use at the moment
 *
 * @param  map  pointer to the map
 *
 * @returns current number of unused buckets
 *//*---------------------------------------------------------------*/

size_t
kv_map_empty_bucket_count(const kv_map_t * map)
{
    node_t ** bp = map->buckets;
    size_t cnt = 0;

    for (size_t pos = 0; pos < map->bucket_cnt; pos++)
        cnt += bp[pos] == NULL;

    return cnt;
}


/*---------------------------------------------------------------*//*
 * @brief Returns the number hash collisions (i.e. the number of
 *        nodes not being the first in a bucket)
 *
 * @param  map  pointer to the map
 *
 * @returns current number of collisions
 *//*---------------------------------------------------------------*/

size_t
kv_map_collision_count(const kv_map_t * map)
{
    if (! map)
        return 0;

    size_t coll_cnt = 0;
    kv_map_iter_t * iter = & KV_MAP_ITER_INIT(map);

    node_t * node = iter->node; inc_iter(iter);
    while (node) {
        if (node != *bucket_ptr(map, node->hash))
            coll_cnt++;
        inc_iter(iter);
        node = iter_to_node(iter);
    }

    return coll_cnt;
}


/*---------------------------------------------------------------*//*
 * @brief Returns the amount of allocated memory currently in  use
 *        for the map
 *
 * @param  map  pointer to the map
 *
 * @returns number of bytes of memory allocated on the heap
 *//*---------------------------------------------------------------*/

size_t
kv_map_mem_size(const kv_map_t * map)
{
    return map->mem_size;
}


/*
 * Local variables:
 * tab-width: 4
 * indent-tabs-mode: nil
 * End:
 */
