/*
 *  Copyright (C) 2020 Jens Thoms Toerring <jt@toerring.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 *  The program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef KV_MAP_H__
#define KV_MAP_H__


#include <stddef.h>
#include <stdbool.h>
#include <stdalign.h>
#include <stdint.h>


typedef struct {
    void     * buckets;                // array of buckets
    size_t     bucket_cnt;             // number of buckets
    size_t     node_cnt;               // number of nodes
    size_t     obj_size;               // aize of the stored objects
    size_t     voffset;                // offset of object from node start addr.
    size_t     koffset;                // offset of key from node start addr.
    size_t     mem_size;               // number of bytes allocated on heap
    uint16_t   align;                  // alignment required for object
    bool       auto_create;            // is auto-create mode enabled?
    int        (* init)(const char *, void *);  // object "constructor"
    void       (* del)(const char *, void *);   // object "destructor"
} kv_map_t;


typedef struct {
    const kv_map_t * map;              // map the iterator is for
    void           * node;             // node iterator is currently set to
} kv_map_iter_t;


#define KV_MAP_INIT(T, MODE, INIT, DEL)                              \
(kv_map_t) {.buckets     = NULL,                                   \
            .bucket_cnt  = 0,                                      \
            .node_cnt    = 00,                                     \
            .mem_size    = 0,                                      \
            .obj_size    = sizeof(T),                              \
            .voffset     = 0,                                      \
            .koffset     = 0,                                      \
            .align       = alignof(T),                             \
            .auto_create = MODE,                                   \
            .init        = (int (*)(const char *, void *)) INIT,   \
            .del         = (void (*)(const char *, void *)) DEL}


/* Note: setting the 'map' and 'node' member to the same address means
 * that the iterator is primed for starting at the first node. */

#define KV_MAP_ITER_INIT(MAP_P)                                     \
(kv_map_iter_t) {.map  = MAP_P,                                     \
                 .node = (void *) MAP_P}


void * kv_map_add(kv_map_t * map, const char * key);

void kv_map_remove(kv_map_t * map, const char * key);

bool kv_map_exists(kv_map_t * map, const char * key);

void * kv_map_get(kv_map_t * map, const char * key);

const char * kv_map_next_key(kv_map_iter_t * iter);

void * kv_map_next_value(kv_map_iter_t * iter);

bool kv_map_next(kv_map_iter_t  * iter, const char ** key, void ** value);

void kv_map_clear(kv_map_t * map);

size_t kv_map_count(const kv_map_t * map);

size_t kv_map_bucket_count(const kv_map_t * map);

size_t kv_map_empty_bucket_count(const kv_map_t * map);

size_t kv_map_collision_count(const kv_map_t * map);

size_t kv_map_mem_size(const kv_map_t * map);

#endif


/*
 * Local variables:
 * tab-width: 4
 * indent-tabs-mode: nil
 * End:
 */
